"""x window related functions"""
from system.xlib.xinput import xinput
from system.xlib.xmsgok import xmsgok
from system.xlib.xyesno import xyesno
from system.xlib.xnotify import xnotify
from system.xlib.xgetpass import xgetpass
